from .DataSet import DataSet
from .. import log

from torch.utils.data import Dataset, DataLoader
from torchvision import transforms
import os
from PIL import Image
import numpy as np
from scipy.io import loadmat
import torch


class Flowers:
    def __init__(self, root, train, transform):
        img_dir = os.path.join(root, '102flowers/jpg')
        img_path = []
        if not os.path.exists(img_dir):
            raise ValueError(f"{img_dir}\n未找到 102flowers/jpg 文件夹，检查是否路径错误")
        else:
            for img in os.listdir(img_dir):
                img_path.append(os.path.join(img_dir, img))
        img_path.sort()
        img_path = np.array(img_path)
        assert(len(img_path) == 8189)

        labels = loadmat(os.path.join(root, 'imagelabels.mat'))
        labels = labels['labels'][0] - 1

        setid = loadmat(os.path.join(root, 'setid.mat'))
        testid = np.append(setid['trnid'][0] - 1, setid['valid'][0] - 1)
        trainid = setid['tstid'][0] - 1

        if train:
            img_path = img_path[trainid]
            self.labels = labels[trainid]
        else:
            img_path = img_path[testid]
            self.labels = labels[testid]

        images = []
        for img in img_path:
            image = Image.open(img)
            image = image.resize((256, 256), Image.ANTIALIAS)
            image = np.array(image)
            images.append(image)
        self.images = np.array(images)

        self.transform = transform

    def __getitem__(self, index):
        target = self.transform(self.images[index])
        label = torch.tensor(self.labels[index]).long()
        return target, label

    def __len__(self):
        return len(self.labels)


@DataSet("102flowers", ("flowers",))
def flowers(train=True, transform=None, download=False, root='./data'):
    dataset = Flowers(root=root, train=train, transform=transform)

    if train:
        log.info(f"load train dataset")
    else:
        log.info(f"load test dataset")
    log.info(f"dataset size: {len(dataset.images)}")
    log.info(f"image size: {dataset.images[0].shape}")
    log.info(f"n_channels: 3")
    log.info(f"classes: 102")
    return dataset



from .DataSet import DataSet
from . import CIFAR
from . import MNIST
from . import SVHN
from . import flowers


def get_dataset(name: str):
    return DataSet.get_dataset(name)


def list_datasets():
    return DataSet.list_datasets()


from .GradCAM import GradCAM
from .GBP import GBP
from .LIME import LIME
from .GuidedGradCAM import GuidedGradCAM
from .SaliencyMap import SaliencyMap
from .Sensitivity import Sensitivity
from .SmoothGrad import SmoothGrad
from .IntegratedGradient import IntegratedGradient
from .Algorithm import Algorithm



import torch
from torch import nn
from torchvision import transforms
import numpy as np
import cv2
from tqdm import trange

from .Algorithm import Algorithm
from .utils import gen_mask_image


class Sensitivity(Algorithm):
    """
    a typical kind of sensitivity analysis
    Taken from the paper https://arxiv.org/abs/1704.03296

    The goal is to solve for a mask that explains why did the network output a score for a certain category.
    We create a low resolution (28x28) mask, and use it to perturb the input image to a deep learning network.
    The perturbation combines a blurred version of the image, the regular image, and the up-sampled mask.
    Wherever the mask contains low values, the input image will become more blurry.

    We want to optimize for the next properties:
    (loss )1.When using the mask to blend the input image and it's blurred versions,
             the score of the target category should drop significantly. The evidence of the category should be removed!
    (loss1)2.The mask should be sparse. Ideally the mask should be the minimal possible mask to drop the category score.
             This translates to a L1(1 - mask) term in the cost function.
    (loss2)3.The mask should be smooth. This translates to a total variation regularization in the cost function.
    (loss3)4.The mask shouldn't over-fit the network. Since the network activations might contain a lot of noise,
             it can be easy for the mask to just learn random values that cause the score to drop without being visually coherent.
             In addition to the other terms, this translates to solving for a lower resolution 28x28 mask.
    """

    def __init__(self, model: nn.Module,
                 image: np.ndarray,
                 lr: int = 0.1,
                 max_iters: int = 500,
                 l1_coeff: int = 0.01,
                 l2_coeff: int = 0.2,
                 l2_beta: int = 3,
                 transform: transforms = None,
                 gpu: bool = True):
        """
        :param model: pytorch model
        :param image: raw image data, only support one image once,
                      if it's 3 channel, you have better to change it to RGB mode
        :param lr: learning rate of generator
        :param max_iters: step times
        :param l1_coeff: the coefficient of loss1
        :param l2_coeff: the coefficient of loss2
        :param l2_beta: the coefficient of regularization
        :param transform: the same as your model's transform
        :param gpu: whether use cuda or not
        """
        super().__init__(model, image, transform, gpu)
        self.lr = lr
        self.max_iters = max_iters
        self.l1_coeff = l1_coeff
        self.l2_coeff = l2_coeff
        self.l2_beta = l2_beta

    def forward(self):
        img = self.transform(self.image).unsqueeze(0).to(self.device)

        blurred_image = np.float32(cv2.medianBlur(self.image, 11)) / 255
        blurred_img = self.transform(blurred_image).unsqueeze(0).to(self.device)

        mask = torch.ones([1, 1, 28, 28], dtype=torch.float32).to(self.device).requires_grad_(True)
        upsample = torch.nn.UpsamplingBilinear2d(size=self.image.shape[:2]).to(self.device)
        optimizer = torch.optim.Adam([mask], lr=self.lr)

        target = self.model(img).argmax(1).cpu().detach().numpy()[0]

        for i in trange(self.max_iters):
            upsampled_mask = upsample(mask)
            upsampled_mask = upsampled_mask.expand(1, 3, upsampled_mask.size(2), upsampled_mask.size(3))
            perturbated_input = img.mul(upsampled_mask) + blurred_img.mul(1 - upsampled_mask)
            noise = torch.normal(mean=0, std=0.2, size=(3, 256, 256)).to(self.device).requires_grad_(True)
            perturbated_input = perturbated_input + noise

            output = torch.nn.Softmax(dim=1)(self.model(perturbated_input))
            loss1 = self.l1_coeff * torch.mean(torch.abs(1 - mask))
            loss2 = self.l2_coeff * self.l2_norm(mask)
            loss3 = output[0, target]
            loss = loss1 + loss2 + loss3

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            mask.data.clamp_(0, 1)

        mask = upsample(mask)
        self.mask = mask.cpu().detach().numpy()[0][0]

        self.masked_image = gen_mask_image(self.mask, self.image)

        return self.mask

    def l2_norm(self, input):
        img = input[0, 0, :]
        row_grad = torch.mean(torch.abs((img[:-1, :] - img[1:, :])).pow(self.l2_beta))
        col_grad = torch.mean(torch.abs((img[:, :-1] - img[:, 1:])).pow(self.l2_beta))
        return row_grad + col_grad



import torch
from torch import nn
from torchvision import transforms
import numpy as np


class Algorithm:
    """
        general algorithm framework
    """
    def __init__(self, model: nn.Module,
                 image: np.ndarray,
                 transform: transforms = None,
                 gpu: bool = True):
        """
        :param model: pytorch model
        :param image: raw image data, only support one image once,
                      if it's 3 channel, you have better to change it to RGB mode
        :param transform: the same as your model's transform
        :param gpu: whether use cuda or not
        """
        self.model = model
        self.model.eval()
        self.image = image
        self.transform = transform
        self.device = torch.device('cuda' if gpu and torch.cuda.is_available() else 'cpu')
        self.mask = None
        self.masked_image = None

    def forward(self):
        pass



import torch
from torch.nn import Module
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm
import sys

from .. import log, Metrics


def train(model: Module, train_loader: DataLoader,
          test_loader: DataLoader, criterion, optimizer, scheduler,
          epochs: int, metrics: Metrics,
          device, tensorboard=None, test_step=10):
    for epoch in range(1, epochs + 1):
        log.debug(f"Epoch: {epoch}")
        loader = tqdm(train_loader, desc=f"| training [{epoch}|{epochs}]", file=sys.stdout)
        avg_loss = _train_step(model, loader, criterion, optimizer, scheduler, metrics, device)
        if tensorboard:
            _tensorboard_plot(tensorboard, "train", avg_loss, metrics, epoch)
        if test_loader and epoch % test_step == 0:
            loader = tqdm(test_loader, desc=f"| testing [{epoch}|{epochs}]", file=sys.stdout)
            avg_loss = _test_step(model, loader, criterion, metrics, device)
            if tensorboard:
                _tensorboard_plot(tensorboard, "test", avg_loss, metrics, epoch)


def _train_step(model: Module, loader:tqdm, criterion,
                optimizer, scheduler, metrics, device):
    model.train()
    model.to(device)
    metrics.reset()
    Loss = 0
    count = 0

    for batch in loader:
        data, target = batch[0].to(device), batch[1].to(device)
        if len(target.size()) > 1:
            target = target.argmax(1)

        output = model(data)
        loss = criterion(output, target)
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        Loss += loss
        count += 1
        metrics(output, target)
        tqdm.set_postfix(loader,
                         {
                             "loss": f"{Loss / count:.6f}",
                             "acc": f"{getattr(metrics, 'accuracy'):.4%}"
                         })
    if scheduler:
        scheduler.step()
    return Loss / count


def _test_step(model: Module, loader: tqdm, criterion,
               metrics, device):
    model.eval()
    model.to(device)
    metrics.reset()
    Loss = 0
    count = 0
    with torch.no_grad():
        for batch in loader:
            data, target = batch[0].to(device), batch[1].to(device)
            if len(target.size()) > 1:
                target = target.argmax(1)

            output = model(data)
            loss = criterion(output, target)

            Loss += loss
            count += 1
            metrics(output, target)
            tqdm.set_postfix(loader,
                             {
                                 "loss": f"{Loss / count:.6f}",
                                 "acc": f"{getattr(metrics, 'accuracy'):.4%}"
                             })
    return Loss / count


def _tensorboard_plot(tensorboard: SummaryWriter, prefix: str,
                      avg_loss: float, metrics: Metrics, epoch: int):
    log.debug(f"Loss: {avg_loss}")
    tensorboard.add_scalar(f"{prefix}_loss", avg_loss, epoch)
    names = metrics.get_names()
    for name in names:
        if hasattr(metrics, name):
            log.debug(f"{name}: {getattr(metrics, name)}")
            tensorboard.add_scalar(f"{prefix}_{name}", getattr(metrics, name), epoch)


import torch
from torch import nn
from torchvision import transforms
import numpy as np

from .Algorithm import Algorithm


class SaliencyMap(Algorithm):
    """
        Saliency Map
        Taken from the paper https://arxiv.org/abs/1312.6034
    """

    def __init__(self, model: nn.Module,
                 image: np.ndarray,
                 transform: transforms = None,
                 gpu: bool = True):
        """
        :param model: pytorch model
        :param image: raw image data, only support one image once,
                      if it's 3 channel, you have better to change it to RGB mode
        :param transform: the same as your model's transform
        :param gpu: whether use cuda or not
        """
        super().__init__(model, image, transform, gpu)

    def forward(self):
        img = self.transform(self.image).unsqueeze(0).to(self.device)
        img.requires_grad = True

        output = self.model(img)
        self.model.zero_grad()
        index = output.argmax(1).cpu().detach().numpy()[0]
        output[0][index].backward()

        self.mask, _ = torch.max(abs(img.grad[0].detach()), dim=0)
        self.mask = self.mask.cpu().numpy()

        return self.mask



import numpy as np
from torch import nn
from torchvision import transforms
import cv2

from .Algorithm import Algorithm
from .utils import gen_mask_image


class GradCAM(Algorithm):
    """
        Grad CAM
        Taken from the paper https://arxiv.org/abs/1610.02391
    """

    def __init__(self, model: nn.Module,
                 image: np.ndarray,
                 last_conv: nn.Conv2d,
                 transform: transforms = None,
                 gpu: bool = True):
        """
        :param model: pytorch model
        :param image: raw image data, only support one image once,
                      if it's 3 channel, you have better to change it to RGB mode
        :param transform: the same as your model's transform
        :param gpu: whether use cuda or not
        """
        super().__init__(model, image, transform, gpu)

        self.last_conv = last_conv
        self.grad_block = []
        self.feature_map_block = []

        self.register_hooks()

    def forward(self):
        img = self.transform(self.image).unsqueeze(0).to(self.device)

        output = self.model(img)
        self.model.zero_grad()
        index = output.argmax(1).cpu().detach().numpy()[0]
        output[0, index].backward()

        grads = self.grad_block[0].cpu().data.numpy()[0]
        feature_maps = self.feature_map_block[0].cpu().data.numpy()[0]

        img = img[0].cpu().numpy().transpose(1, 2, 0)
        cam = np.zeros(feature_maps.shape[1:], dtype=np.float32)

        H, W, C = img.shape
        grads = grads.reshape([grads.shape[0], -1])
        weights = np.mean(grads, axis=1)
        for i, w in enumerate(weights):
            cam += w * feature_maps[i, :, :]
        cam = np.maximum(cam, 0)
        cam = cam / cam.max()
        cam = cv2.resize(cam, (W, H))
        self.mask = cam

        self.masked_image = gen_mask_image(self.mask, self.image)

        return self.mask

    def register_hooks(self):
        self.last_conv.register_forward_hook(self.forward_hook)
        self.last_conv.register_backward_hook(self.backward_hook)

    def backward_hook(self, module, grad_in, grad_out):
        self.grad_block.append(grad_in[0].detach())

    def forward_hook(self, module, input, output):
        self.feature_map_block.append(output)



import matplotlib.pyplot as plt
from typing import List


def subplot(params: List[dict], cmap=None):
    plt.figure()
    rows = len(params)
    cols = max([len(i) for i in params])
    for r in range(rows):
        index = 0
        for key, value in params[r].items():
            index += 1
            plt.subplot(rows, cols, r*cols+index)
            plt.axis('off')
            if r == 0:
                plt.title(key)
            plt.imshow(value, cmap=cmap)
    plt.show()




import numpy as np
import torch
from torch import nn
from torchvision import transforms
from tqdm import trange

from .Algorithm import Algorithm
from ..logger import log


class SmoothGrad(Algorithm):
    """
        Saliency Map
        Taken from the paper https://arxiv.org/abs/1312.6034
    """

    def __init__(self, model: nn.Module,
                 image: np.ndarray,
                 stdev_spread: int = 0.15,
                 n_samples: int = 25,
                 magnitude: bool = False,
                 transform: transforms = None,
                 gpu: bool = True):
        """
        :param model: pytorch model
        :param image: raw image data, only support one image once,
                      if it's 3 channel, you have better to change it to RGB mode
        :param stdev_spread: to control the std of Gaussian distribution, generating noise data
        :param n_samples: the times of sampling noise, to reduce error
        :param magnitude: whether to scale the grad by pow or not
        :param transform: the same as your model's transform
        :param gpu: whether use cuda or not
        """
        super().__init__(model, image, transform, gpu)
        self.stdev_spread = stdev_spread
        self.n_samples = n_samples
        self.magnitutde = magnitude

    def forward(self):
        img = self.transform(self.image).unsqueeze(0).to(self.device)
        stdev = self.stdev_spread * (img.max() - img.min())
        total_gradients = np.zeros_like(img.cpu().numpy())
        for i in trange(self.n_samples):
            log.info(f"epoch [{i + 1}|{self.n_samples}]...")
            noise = torch.normal(mean=0, std=stdev, size=img.shape, dtype=torch.float32).to(self.device)
            img_plus_noise = img + noise
            img_plus_noise.requires_grad = True

            output = self.model(img_plus_noise)
            index = output.argmax(1).cpu().detach().numpy()[0]
            output[0, index].backward()

            grad = img_plus_noise.grad.cpu().detach().numpy()

            if self.magnitutde:
                total_gradients += (grad * grad)
            else:
                total_gradients += grad

        log.info("finish perturbation.")
        self.mask = total_gradients[0, :, :, :] / self.n_samples
        self.mask = (self.mask - self.mask.min()) / (self.mask.max() - self.mask.min())
        self.mask = self.mask.transpose(1, 2, 0)

        return self.mask



from typing import Tuple
from ..logger import log


class Models:
    _models = {}

    def __init__(self, name: str, alias: Tuple[str, ...] = ()):
        self.name = name
        self.alias = alias

    def __call__(self, function):
        if self._models.get(self.name):
            log.warn(f"模型{self.name}已存在")
        else:
            self._models[self.name] = function
        for alias in self.alias:
            if self._models.get(alias):
                log.warn(f"模型别名{alias}已存在")
            else:
                self._models[alias] = function

    @staticmethod
    def get_model(name: str):
        if Models._models.get(name):
            return Models._models.get(name)
        else:
            log.critical(f"模型{name}不存在")


from torchvision.models import resnet18, resnet34, resnet50, resnet101, resnet152
from .Models import Models
import torch.nn as nn


def ResNet(model, in_channels=3, classes: int = 1000, pretrained: bool = False, **kwargs):
    resnet = model(pretrained, **kwargs)
    resnet.conv1 = nn.Conv2d(in_channels=in_channels, out_channels=64,
                             kernel_size=7, stride=2, padding=3, bias=False)
    in_features = resnet.fc.in_features
    resnet.fc = nn.Linear(in_features, classes)
    return resnet


@Models("resnet18")
def ResNet18(in_channels=3, classes: int = 102, pretrained: bool = False, **kwargs):
    return ResNet(resnet18, in_channels, classes, pretrained, **kwargs)


@Models("resnet34")
def ResNet34(in_channels=3, classes: int = 1000, pretrained: bool = False, **kwargs):
    return ResNet(resnet34, in_channels, classes, pretrained, **kwargs)


@Models("resnet50")
def ResNet50(in_channels=3, classes: int = 1000, pretrained: bool = False, **kwargs):
    return ResNet(resnet50, in_channels, classes, pretrained, **kwargs)


@Models("resnet101")
def ResNet101(in_channels=3, classes: int = 1000, pretrained: bool = False, **kwargs):
    return ResNet(resnet101, in_channels, classes, pretrained, **kwargs)


@Models("resnet152")
def ResNet152(in_channels=3, classes: int = 1000, pretrained: bool = False, **kwargs):
    return ResNet(resnet152, in_channels, classes, pretrained, **kwargs)


from .logger import log, get_logger
from .datasets import get_dataset, list_datasets
from .model import get_model
from .metrics import ConfuseMatrix, Metrics
from .train import train
from .algorithnm import GradCAM, GBP, LIME, GuidedGradCAM, SaliencyMap, Sensitivity, SmoothGrad, IntegratedGradient
from .visualize import subplot



import logging
import sys
import os
import json
from .ColoredLogger import ColoredFormatter

with open(os.path.join(os.path.dirname(__file__), './logger_config.json'), encoding="UTF-8") as f:
    config = json.loads(f.read())


def init_logger():
    # 创建logger记录器
    logger = logging.getLogger(config.get("logger_name"))
    logger.setLevel(logging.DEBUG)

    # 创建控制台处理器
    sh = logging.StreamHandler()
    sh.setLevel(logging.DEBUG)
    sh.setStream(sys.stdout)

    # 创建formatter格式化器
    formatter = ColoredFormatter(fmt="[%(levelname)s]"
                                     "%(asctime)s"
                                     # "[%(name)s]"
                                     # "(%(filename)s,%(lineno)s)"
                                     " | %(message)s",
                                 use_color=config.get("use_color"),
                                 datafmt="%Y-%m-%d %H:%M:%S"
                                 )

    # 将formatter添加到控制台处理器
    sh.setFormatter(formatter)

    # 将控制台处理器添加到logger
    logger.addHandler(sh)


def get_logger():
    return logging.getLogger(config.get("logger_name"))



from .DataSet import DataSet
from torchvision.datasets import SVHN
from .. import log


@DataSet("SVHN", ("svhn",))
def mnist(train=True, transform=None, download=False, root='./data'):
    if train:
        split = 'train'
    else:
        split = 'test'
    dataset = SVHN(split=split, transform=transform, download=download, root=root)
    if train:
        log.info(f"load train dataset")
    else:
        log.info(f"load test dataset")
    log.info(f"dataset size: {len(dataset.labels)}")
    log.info(f"image size: {dataset.data[0].shape}")
    log.info(f"n_channels: 3")
    log.info(f"labels: {set(dataset.labels)}")
    return dataset


import numpy as np
import cv2


def gen_mask_image(mask: np.ndarray, image: np.ndarray,
                   mask_trans=0.5, image_trans=0.5, colormap=cv2.COLORMAP_JET):
    if len(image.shape) != 3:
        raise RuntimeError("image维数错误")
    if len(mask.shape) == 2:
        heatmap = cv2.applyColorMap(np.uint8(255 * mask), colormap)
        heatmap = heatmap[:, :, ::-1]
        masked_image = cv2.addWeighted(image, image_trans, heatmap, mask_trans, 0)
        return masked_image
    elif len(mask.shape) == 3:
        return cv2.addWeighted(image, image_trans, mask, mask_trans, 0)
    else:
        raise RuntimeError("mask维数错误")


from torch import nn, clamp

from .GradCAM import GradCAM
from .utils import gen_mask_image


class GuidedGradCAM(GradCAM):
    """
        GradCAM + Guided BP
        a little change
    """
    def __init__(self, model, image, last_conv, transform=None, gpu=True):
        super().__init__(model, image, last_conv, transform, gpu)

    def plot(self):
        self.masked_image = gen_mask_image(self.mask, self.image)
        params = {
            "origin image": self.image,
            "GuidedGradCAM": self.mask,
            "masked image": self.masked_image
        }
        return params

    def register_hooks(self):
        self.last_conv.register_forward_hook(self.forward_hook)
        self.last_conv.register_backward_hook(self.backward_hook)

        for module in self.model.children():
            if isinstance(module, nn.ReLU):
                module.register_backward_hook(self.backward_relu_hook)

    def backward_relu_hook(self, module, grad_in, grad_out):
        return (clamp(grad_in[0], min=0), )





import numpy as np
from torch import nn
from torchvision import transforms
from tqdm import trange

from .Algorithm import Algorithm


class IntegratedGradient(Algorithm):
    """
        Integrated Gradient
        Taken from the paper https://arxiv.org/abs/1703.01365
    """

    def __init__(self, model: nn.Module,
                 image: np.ndarray,
                 n_samples: int = 10,
                 steps: int = 50,
                 target_lable: int = None,
                 transform: transforms = None,
                 gpu: bool = True):
        """
        :param model: pytorch model
        :param image: raw image data, only support one image once,
                      if it's 3 channel, you have better to change it to RGB mode
        :param n_samples: the times of sampling a baseline, to reduce error
        :param steps: step size in approximate integration of rectangular method
        :param target_lable: the lable corresponding to the image
        :param transform: the same as your model's transform
        :param gpu: whether use cuda or not
        """
        super().__init__(model, image, transform, gpu)

        if target_lable:
            self.target_lable = target_lable
        else:
            img = self.transform(self.image).unsqueeze(0).to(self.device)
            self.target_lable = self.model(img).argmax(1).cpu().detach().numpy()[0]

        self.n_samples = n_samples
        self.steps = steps

    def forward(self):
        all_integrated_grads = []
        for i in trange(self.n_samples):
            baseline = 255.0 * np.random.random(self.image.shape)
            scaled_images = [baseline + (float(i) / self.steps) * (self.image - baseline) for i in
                             range(0, self.steps + 1)]

            gradients = []
            for x in scaled_images:
                x = self.transform(x).float().unsqueeze(0).to(self.device).requires_grad_(True)

                output = self.model(x)
                self.model.zero_grad()
                output[0, self.target_lable].backward()

                gradient = x.grad.detach().cpu().numpy()[0]
                gradients.append(gradient)

            avg_grads = np.average(gradients[:-1], axis=0)
            delta_img = (self.transform(self.image) - self.transform(baseline)).numpy()
            int_grad = avg_grads * delta_img
            int_grad = (int_grad - int_grad.min()) / (int_grad.max() - int_grad.min())
            all_integrated_grads.append(int_grad)

        avg_integrated_grad = np.average(np.array(all_integrated_grads), axis=0)
        self.mask = avg_integrated_grad.transpose(1, 2, 0)

        return self.mask



from lime import lime_image
import torch
import numpy as np
from torch import nn
from torchvision import transforms

from .Algorithm import Algorithm
from .utils import gen_mask_image


class LIME(Algorithm):
    """
        Local Interpretable Model-agnostic Explanations
        Taken from the paper https://arxiv.org/abs/1602.04938v3
    """

    def __init__(self, model: nn.Module,
                 image: np.ndarray,
                 label: int = None,
                 transform: transforms = None,
                 gpu: bool = True):
        """
        :param model: pytorch model
        :param image: raw image data, only support one image once,
                      if it's 3 channel, you have better to change it to RGB mode
        :param transform: the same as your model's transform
        :param gpu: whether use cuda or not
        """
        super().__init__(model, image, transform, gpu)

        if label:
            self.label = label
        else:
            img = self.transform(self.image).unsqueeze(0).to(self.device)
            output = self.model(img)
            label = output.argmax(1).cpu().detach().numpy()[0]
            self.label = label

    def forward(self):
        explainer = lime_image.LimeImageExplainer()
        result = explainer.explain_instance(self.image, self.predict_fn, labels=self.label, num_samples=1000)
        _, mask = result.get_image_and_mask(result.top_labels[0], positive_only=True, num_features=5, hide_rest=False)
        self.mask = mask

        self.masked_image = gen_mask_image(self.mask, self.image)

        return self.mask

    def predict_fn(self, x):
        x = torch.tensor(np.array([self.transform(i).numpy() for i in x])).to(self.device)
        pred = self.model(x)
        return pred.cpu().detach().numpy()



from typing import Tuple
from .. import log


class DataSet:
    _datasets = {}

    def __init__(self, name: str, alias: Tuple[str, ...] = ()):
        self.name = name
        self.alias = alias

    def __call__(self, function):
        if self._datasets.get(self.name):
            log.warn(f"数据集{self.name}已存在")
        else:
            self._datasets[self.name] = function
        for alias in self.alias:
            if self._datasets.get(alias):
                log.warn(f"数据集别名{alias}已存在")
            else:
                self._datasets[alias] = function

    @staticmethod
    def get_dataset(name: str):
        if DataSet._datasets.get(name):
            return DataSet._datasets.get(name)
        else:
            log.critical(f"数据集{name}不存在")

    @staticmethod
    def list_datasets():
        for k, v in DataSet._datasets.items():
            print(f"{k}: {v}")



import logging
import os
import json


class ColoredFormatter(logging.Formatter):
    with open(os.path.join(os.path.dirname(__file__), "color_config.json"), encoding='UTF-8') as f:
        config = json.loads(f.read())

    def __init__(self, fmt, use_color=True, datafmt=None, **kwargs):
        super().__init__(fmt, datafmt, **kwargs)
        self.use_color = use_color

    def format(self, record):
        if self.use_color:
            levelname = record.levelname
            mode = self.config['MODE']
            foreground = self.config['FOREGROUND']
            background = self.config['BACKGROUND']
            prefix = f"\033[{mode[levelname]};{foreground[levelname]};{background[levelname]}m"
            suffix = f"\033[0m"
            record.levelname = f"{prefix}{levelname}{suffix}"
        return super().format(record)



from typing import Tuple
import numpy as np
from torch import Tensor
from .. import log
from .Metrics import Metrics


class ConfuseMatrix(Metrics):
    def __init__(self, classes: int, names: Tuple[str, ...] = (), *args, **kwargs):
        super().__init__(names)
        self.classes = classes
        self.confuse_matrix = np.zeros((classes, classes), dtype=np.int)

    def forward(self, output: Tensor, target: Tensor):
        predicts = output.argmax(1).cpu().numpy()
        target = target.cpu().numpy()
        for i, j in zip(predicts, target):
            self.confuse_matrix[i, j] += 1

    def reset(self):
        self.confuse_matrix = np.zeros((self.classes, self.classes), dtype=np.int)

    @property
    def accuracy(self):
        correct = 0
        for n in range(self.classes):
            correct += self.confuse_matrix[n, n]
        Accuracy = correct / self.confuse_matrix.sum()
        return Accuracy

    @property
    def recall(self):
        if self.classes != 2:
            log.critical("不是二分类问题，请调用macro_recall")
            raise Exception()
        AT = self.confuse_matrix.sum(axis=0)[1]
        TP = self.confuse_matrix[1, 1]
        Recall = TP / AT
        return Recall

    @property
    def precision(self):
        if self.classes != 2:
            log.critical("不是二分类问题，请调用macro_precision")
            raise Exception()
        PT = self.confuse_matrix.sum(axis=1)[1]
        TP = self.confuse_matrix[1, 1]
        Precision = TP / PT
        return Precision

    @property
    def f1_score(self):
        Recall = self.recall
        Precision = self.precision
        F1_score = 2 * Precision * Recall / (Precision + Recall)
        return F1_score

    @property
    def micro_recall(self):
        AT = self.confuse_matrix.sum(axis=0)
        TP = np.array([self.confuse_matrix[i, i] for i in range(self.classes)])
        MicroRecall = AT.sum() / TP.sum()
        return MicroRecall

    @property
    def macro_recall(self):
        AT = self.confuse_matrix.sum(axis=0)
        TP = np.array([self.confuse_matrix[i, i] for i in range(self.classes)])
        temp = TP / AT
        log.debug(temp)
        MacroRecall = temp.mean()
        return MacroRecall

    @property
    def micro_precision(self):
        PT = self.confuse_matrix.sum(axis=1)
        TP = np.array([self.confuse_matrix[i, i] for i in range(self.classes)])
        MicroPrecision = TP.sum() / PT.sum()
        return MicroPrecision

    @property
    def macro_precision(self):
        PT = self.confuse_matrix.sum(axis=1)
        TP = np.array([self.confuse_matrix[i, i] for i in range(self.classes)])
        temp = TP / PT
        log.debug(temp)
        MacroPrecision = temp.mean()
        return MacroPrecision




from torch import nn, clamp
import numpy as np
from torchvision import transforms
import cv2

from .Algorithm import Algorithm


class GBP(Algorithm):
    """
        Guided Backpropagation
        Taken from the paper https://arxiv.org/abs/1412.6806
    """

    def __init__(self, model: nn.Module,
                 image: np.ndarray,
                 transform: transforms = None,
                 gpu: bool = True):
        """
        :param model: pytorch model
        :param image: raw image data, only support one image once,
                      if it's 3 channel, you have better to change it to RGB mode
        :param transform: the same as your model's transform
        :param gpu: whether use cuda or not
        """
        super().__init__(model, image, transform, gpu)

        self.image_reconstruction = []
        self.activation_maps = []

        self.register_hooks()

    def forward(self):
        img = self.transform(self.image).unsqueeze(0).to(self.device)
        img.requires_grad = True

        output = self.model(img)
        self.model.zero_grad()
        index = output.argmax(1).cpu().detach().numpy()[0]
        output[0][index].backward()

        result = img.grad[0].detach().cpu().numpy().transpose(1, 2, 0)
        result = (result - result.mean()) / result.std()
        result *= 0.1
        result += 0.5
        H, W, C = result.shape
        result = cv2.resize(result, (W, H))
        result = result.clip(0, 1)
        self.mask = result

        return self.mask

    def register_hooks(self):
        for name, module in self.model.named_children():
            if isinstance(module, nn.ReLU):
                module.register_forward_hook(self.forward_hook)
                module.register_backward_hook(self.backward_hook)

    def forward_hook(self, module, input, output):
        self.activation_maps.append(output)

    def backward_hook(self, module, grad_in, grad_out):
        return (clamp(grad_in[0], min=0.0),)



import setuptools
with open("README.md", "r", encoding='utf-8') as f:
    long_description = f.read()
setuptools.setup(
    name="vinn",
    version="1.1",
    author="qianzhihao",
    author_email="2018302170027@whu.edu.cn",
    description="Visual Interpretation of Neural Networks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Torato-Taraka/VINN",
    packages=setuptools.find_packages(),
    install_requires=[],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)


from typing import Tuple
from torch import Tensor


class Metrics:
    def __init__(self, names: Tuple[str, ...] = ()):
        self._names = names

    def __call__(self, output: Tensor, target: Tensor):
        self.forward(output, target)

    def forward(self, output: Tensor, target: Tensor):
        pass

    def reset(self):
        pass

    def get_names(self):
        return self._names
    


import torch
from torch.utils.data import DataLoader
from torch.nn import CrossEntropyLoss
from torch.optim import Adam
from torch.optim.lr_scheduler import MultiStepLR
from torch.utils.tensorboard import SummaryWriter

from .. import get_dataset, get_model, train, ConfuseMatrix


def trainer(model_name, dataset_name, classes=1000, transform=None,
            train_size=64, test_size=50, lr=0.001, epochs=100, gpu=True):
    # device
    device = torch.device('cuda' if gpu and torch.cuda.is_available() else 'cpu')

    # dataset
    dataset = get_dataset(dataset_name)
    train_dataset = dataset(train=True, transform=transform, download=True, root='../data')
    test_dataset = dataset(train=False, transform=transform, download=True, root='../data')
    train_loader = DataLoader(train_dataset, batch_size=train_size, shuffle=True)
    test_loader = DataLoader(test_dataset, batch_size=test_size, shuffle=False)

    # model
    model = get_model(model_name)().to(device)

    # train
    criterion = CrossEntropyLoss()
    optimizer = Adam(model.parameters(), lr=lr)

    # scheduler = MultiStepLR(optimizer, milestones=[30, 60, 90], gamma=0.1)
    scheduler = None

    metrics = ConfuseMatrix(classes=classes, names=("accuracy",))

    # tensorboard = SummaryWriter(comment=f"{model_name}_{dataset_name}")
    tensorboard = None

    train(model=model, train_loader=train_loader, test_loader=test_loader,
          criterion=criterion, optimizer=optimizer,
          scheduler=scheduler, epochs=epochs,
          metrics=metrics, device=device, tensorboard=tensorboard)

    torch.save(model, f"./trained/{model_name}_{dataset_name}.pth")



from .DataSet import DataSet
from torchvision.datasets import MNIST
from .. import log


@DataSet("MNIST", ("mnist",))
def mnist(train=True, transform=None, download=False, root='./data'):
    dataset = MNIST(train=train, transform=transform, download=download, root=root)
    if train:
        log.info(f"load train dataset")
    else:
        log.info(f"load test dataset")
    log.info(f"dataset size: {len(dataset.targets)}")
    log.info(f"image size: {dataset.data[0].shape}")
    log.info(f"n_channels: 1")
    log.info(f"{dataset.classes}")
    return dataset